package gochem

import (
	"bufio"
	"fmt"
	"os"
	"strings"
)

const (
	maxAtoms      = 1000
	maxNeighbours = 4
)

type Atom struct {
	X, Y, Z       float64
	AtomType      string
	Idx           int
	Neighbours    []*Atom
	BondOrders    []int
	BondsDrawn    bool
	NumNeighbours int
}

type Molecule struct {
	Atoms    []*Atom
	NumAtoms int
}

func NewAtom() *Atom {
	return &Atom{
		Neighbours: make([]*Atom, 0, maxNeighbours),
		BondOrders: make([]int, 0, maxNeighbours),
	}
}

func ParseSDF(filename string) (*Molecule, error) {
	file, err := os.Open(filename)
	if err != nil {
		return nil, fmt.Errorf("error opening file %s: %v", filename, err)
	}
	defer file.Close()

	mol := &Molecule{
		Atoms:    make([]*Atom, 0, maxAtoms),
		NumAtoms: 0,
	}

	scanner := bufio.NewScanner(file)
	for scanner.Scan() {
		line := scanner.Text()

		// Stop when "M  END" is reached
		if strings.HasPrefix(line, "M  END") {
			fmt.Println("Reached END")
			break
		}

		// Skip lines containing "V2000" or "V3000"
		if strings.Contains(line, "V2000") || strings.Contains(line, "V3000") {
			continue
		}

		// Parse atoms (coordinates and atom type)
		if isAtomLine(line) {
			atom := NewAtom()
			fmt.Sscanf(line, "%f %f %f %s", &atom.X, &atom.Y, &atom.Z, &atom.AtomType)
			atom.Idx = mol.NumAtoms + 1
			mol.Atoms = append(mol.Atoms, atom)
			mol.NumAtoms++
		}

		// Parse bond connectivity
		if isBondLine(line) {
			var atomIdx, neighbourIdx, bondOrder int
			fmt.Sscanf(line, "%d %d %d", &atomIdx, &neighbourIdx, &bondOrder)

			addBond(mol.Atoms[atomIdx-1], mol.Atoms[neighbourIdx-1], bondOrder)
		}
	}

	if err := scanner.Err(); err != nil {
		return nil, fmt.Errorf("error reading file %s: %v", filename, err)
	}

	return mol, nil
}

func isAtomLine(line string) bool {
	// Check that 5th and 7th characters are digits and 32nd is a letter
	return len(line) > 31 && isDigit(line[4]) && isDigit(line[6]) && isAlpha(line[31])
}

func isBondLine(line string) bool {
	// Check bond line pattern (digits at expected positions)
	return len(line) > 20 && isDigit(line[2]) && isDigit(line[20])
}

func isDigit(b byte) bool {
	return b >= '0' && b <= '9'
}

func isAlpha(b byte) bool {
	return (b >= 'A' && b <= 'Z') || (b >= 'a' && b <= 'z')
}

func addBond(atom1, atom2 *Atom, bondOrder int) {
	if len(atom1.Neighbours) < maxNeighbours && len(atom2.Neighbours) < maxNeighbours {
		atom1.Neighbours = append(atom1.Neighbours, atom2)
		atom1.BondOrders = append(atom1.BondOrders, bondOrder)
		atom1.NumNeighbours++

		atom2.Neighbours = append(atom2.Neighbours, atom1)
		atom2.BondOrders = append(atom2.BondOrders, bondOrder)
		atom2.NumNeighbours++
	} else {
		fmt.Println("ERROR: MAX_NEIGHBOURS exceeded!")
	}
}

func NormalizeCoordinates(mol *Molecule, width, height int) {
	if mol.NumAtoms == 0 {
		return
	}

	minX, maxX := mol.Atoms[0].X, mol.Atoms[0].X
	minY, maxY := mol.Atoms[0].Y, mol.Atoms[0].Y
	minZ, maxZ := mol.Atoms[0].Z, mol.Atoms[0].Z

	// Find min and max coordinates
	for _, atom := range mol.Atoms[1:] {
		if atom.X < minX {
			minX = atom.X
		}
		if atom.X > maxX {
			maxX = atom.X
		}
		if atom.Y < minY {
			minY = atom.Y
		}
		if atom.Y > maxY {
			maxY = atom.Y
		}
		if atom.Z < minZ {
			minZ = atom.Z
		}
		if atom.Z > maxZ {
			maxZ = atom.Z
		}
	}

	// Scaling factors
	scaleX := float64(width) / (maxX - minX) / 100
	scaleY := float64(height) / (maxY - minY) / 100
	scaleZ := float64(height) / (maxZ - minZ) / 100

	// Normalize coordinates
	for _, atom := range mol.Atoms {
		atom.X = (atom.X - minX) * scaleX
		atom.Y = (atom.Y - minY) * scaleY
		atom.Z = (atom.Z - minZ) * scaleZ
	}
}

